/*************************************************************************
 *  Copyright (C) 2012 by Volker Lanz <vl@fidra.de>                      *
 *  Copyright (C) 2016 by Andrius Štikonas <andrius@stikonas.eu>         *
 *                                                                       *
 *  This program is free software; you can redistribute it and/or        *
 *  modify it under the terms of the GNU General Public License as       *
 *  published by the Free Software Foundation; either version 3 of       *
 *  the License, or (at your option) any later version.                  *
 *                                                                       *
 *  This program is distributed in the hope that it will be useful,      *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *  GNU General Public License for more details.                         *
 *                                                                       *
 *  You should have received a copy of the GNU General Public License    *
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.*
 *************************************************************************/

#include "fs/lvm2_pv.h"

#include "util/externalcommand.h"
#include "util/capacity.h"

#include <QString>

#include <KLocalizedString>

namespace FS
{
FileSystem::CommandSupportType lvm2_pv::m_GetUsed = FileSystem::cmdSupportNone;
FileSystem::CommandSupportType lvm2_pv::m_GetLabel = FileSystem::cmdSupportNone;
FileSystem::CommandSupportType lvm2_pv::m_Create = FileSystem::cmdSupportNone;
FileSystem::CommandSupportType lvm2_pv::m_Grow = FileSystem::cmdSupportNone;
FileSystem::CommandSupportType lvm2_pv::m_Shrink = FileSystem::cmdSupportNone;
FileSystem::CommandSupportType lvm2_pv::m_Move = FileSystem::cmdSupportNone;
FileSystem::CommandSupportType lvm2_pv::m_Check = FileSystem::cmdSupportNone;
FileSystem::CommandSupportType lvm2_pv::m_Copy = FileSystem::cmdSupportNone;
FileSystem::CommandSupportType lvm2_pv::m_Backup = FileSystem::cmdSupportNone;
FileSystem::CommandSupportType lvm2_pv::m_SetLabel = FileSystem::cmdSupportNone;
FileSystem::CommandSupportType lvm2_pv::m_UpdateUUID = FileSystem::cmdSupportNone;
FileSystem::CommandSupportType lvm2_pv::m_GetUUID = FileSystem::cmdSupportNone;

lvm2_pv::lvm2_pv(qint64 firstsector, qint64 lastsector, qint64 sectorsused, const QString& label) :
    FileSystem(firstsector, lastsector, sectorsused, label, FileSystem::Lvm2_PV)
{
}

void lvm2_pv::init()
{
    CommandSupportType lvmFound = findExternal(QStringLiteral("lvm")) ? cmdSupportFileSystem : cmdSupportNone;

    m_Create     = lvmFound;
    m_Check      = lvmFound;
    m_Grow       = lvmFound;
    m_Shrink     = lvmFound;
    m_UpdateUUID = lvmFound;
    m_GetUsed    = lvmFound;

    m_Move = (m_Check != cmdSupportNone) ? cmdSupportCore : cmdSupportNone;

    m_GetLabel = cmdSupportCore;
    m_Backup   = cmdSupportCore;
    m_GetUUID  = cmdSupportCore;

    m_GetLabel = cmdSupportNone;
    m_Copy     = cmdSupportNone; // Copying PV can confuse LVM
}

bool lvm2_pv::supportToolFound() const
{
    return
        m_GetUsed != cmdSupportNone &&
//          m_GetLabel != cmdSupportNone &&
//          m_SetLabel != cmdSupportNone &&
        m_Create != cmdSupportNone &&
        m_Check != cmdSupportNone &&
        m_UpdateUUID != cmdSupportNone &&
        m_Grow != cmdSupportNone &&
        m_Shrink != cmdSupportNone &&
//          m_Copy != cmdSupportNone &&
        m_Move != cmdSupportNone &&
        m_Backup != cmdSupportNone &&
        m_GetUUID != cmdSupportNone;
}

FileSystem::SupportTool lvm2_pv::supportToolName() const
{
    return SupportTool(QStringLiteral("lvm2"), QUrl(QStringLiteral("http://sourceware.org/lvm2/")));
}

qint64 lvm2_pv::maxCapacity() const
{
    return Capacity::unitFactor(Capacity::Byte, Capacity::EiB);
}

qint64 lvm2_pv::readUsedCapacity(const QString& deviceNode) const
{
    QString val = getpvField(QStringLiteral("pv_used"), deviceNode);
    QString metadataOffset = getpvField(QStringLiteral("pe_start"), deviceNode);
    return val.isEmpty() ? -1 : val.toLongLong() + metadataOffset.toLongLong();
}

bool lvm2_pv::check(Report& report, const QString& deviceNode) const
{
    ExternalCommand cmd(report, QStringLiteral("lvm"), { QStringLiteral("pvck"), QStringLiteral("--verbose"), deviceNode });
    return cmd.run(-1) && cmd.exitCode() == 0;
}

bool lvm2_pv::create(Report& report, const QString& deviceNode) const
{
    ExternalCommand cmd(report, QStringLiteral("lvm"), { QStringLiteral("pvcreate"), QStringLiteral("--force"), deviceNode });
    return cmd.run(-1) && cmd.exitCode() == 0;
}

bool lvm2_pv::remove(Report& report, const QString& deviceNode) const
{
//      TODO: check if PV is a member of an exported VG
    ExternalCommand cmd(report, QStringLiteral("lvm"), { QStringLiteral("pvremove"), QStringLiteral("--force"), QStringLiteral("--force"), QStringLiteral("--yes"), deviceNode });
    return cmd.run(-1) && cmd.exitCode() == 0;
}

bool lvm2_pv::resize(Report& report, const QString& deviceNode, qint64 length) const
{
    bool rval = true;

    qint64 metadataOffset = getpvField(QStringLiteral("pe_start"), deviceNode).toLongLong();

    qint64 lastPE = getTotalPE(deviceNode) - 1; // starts from 0
    if (lastPE > 0) { // make sure that the PV is already in a VG
        qint64 targetPE = (length - metadataOffset)/ getPESize(deviceNode) - 1; // starts from 0
        if (targetPE < lastPE) { //shrinking FS
            qint64 firstMovedPE = qMax(targetPE + 1, getAllocatedPE(deviceNode)); // starts from 1
            ExternalCommand moveCmd(report,
                                    QStringLiteral("lvm"), {
                                    QStringLiteral("pvmove"),
                                    QStringLiteral("--alloc"),
                                    QStringLiteral("anywhere"),
                                    deviceNode + QStringLiteral(":") + QString::number(firstMovedPE) + QStringLiteral("-") + QString::number(lastPE),
                                    deviceNode + QStringLiteral(":") + QStringLiteral("0-") + QString::number(firstMovedPE - 1)
                                    });
            rval = moveCmd.run(-1) && (moveCmd.exitCode() == 0 || moveCmd.exitCode() == 5); // FIXME: exit code 5: NO data to move
        }
    }

    ExternalCommand cmd(report, QStringLiteral("lvm"), {
                                QStringLiteral("pvresize"),
                                QStringLiteral("--yes"),
                                QStringLiteral("--setphysicalvolumesize"),
                                QString::number(length) + QStringLiteral("B"),
                                deviceNode });
    return rval && cmd.run(-1) && cmd.exitCode() == 0;
}

bool lvm2_pv::updateUUID(Report& report, const QString& deviceNode) const
{
    ExternalCommand cmd(report, QStringLiteral("lvm"), { QStringLiteral("pvchange"), QStringLiteral("--uuid"), deviceNode });
    return cmd.run(-1) && cmd.exitCode() == 0;
}

QString lvm2_pv::readUUID(const QString& deviceNode) const
{
    return getpvField(QStringLiteral("pv_uuid"), deviceNode);
}

bool lvm2_pv::mount(Report& report, const QString& deviceNode, const QString& mountPoint)
{
    Q_UNUSED(report);
    Q_UNUSED(deviceNode);
    Q_UNUSED(mountPoint);
    return false;
}

bool lvm2_pv::unmount(Report& report, const QString& deviceNode)
{
    Q_UNUSED(deviceNode);
    Q_UNUSED(report);
    return false;
}

bool lvm2_pv::canMount(const QString & deviceNode, const QString & mountPoint) const
{
    Q_UNUSED(mountPoint);
    QString rval = getpvField(QStringLiteral("pv_in_use"), deviceNode); // this field return "used" when in use otherwise empty string
    if (rval.isEmpty()) {
        return true;
    }
    return false;
}

bool lvm2_pv::canUnmount(const QString& deviceNode) const
{
    QString rval = getpvField(QStringLiteral("pv_in_use"), deviceNode);
    if (rval.isEmpty()) {
        return false;
    }
    return true;
}

qint64 lvm2_pv::getTotalPE(const QString& deviceNode)
{
    QString val = getpvField(QStringLiteral("pv_pe_count"), deviceNode);
    return val.isEmpty() ? -1 : val.toLongLong();
}

qint64 lvm2_pv::getTotalPE(const QStringList& deviceNodeList)
{
    qint64 sum = 0;
    for (const auto &deviceNode : deviceNodeList) {
        qint64 totalPE =  getTotalPE(deviceNode);
        if (totalPE < 0) {
            sum = -1;
            break;
        }
        sum += totalPE;
    }
    return sum;
}

qint64 lvm2_pv::getFreePE(const QString& deviceNode)
{
    return getTotalPE(deviceNode) - getAllocatedPE(deviceNode);
}

qint64 lvm2_pv::getFreePE(const QStringList& deviceNodeList)
{
    qint64 sum = 0;
    for (QString deviceNode :deviceNodeList) {
        qint64 freePE = getFreePE(deviceNode);
        if (freePE < 0) {
            sum = -1;
            break;
        }
        sum += freePE;
    }
    return sum;
}

qint64 lvm2_pv::getAllocatedPE(const QString& deviceNode)
{
    QString val = getpvField(QStringLiteral("pv_pe_alloc_count"), deviceNode);
    return val.isEmpty() ? -1 : val.toLongLong();
}

qint64 lvm2_pv::getAllocatedPE(const QStringList& deviceNodeList)
{
    qint64 sum = 0;
    for (QString deviceNode : deviceNodeList) {
        qint64 allocatedPE = getAllocatedPE(deviceNode);
        if (allocatedPE < 0) {
            sum = -1;
            break;
        }
        sum += allocatedPE;
    }
    return sum;
}

qint64 lvm2_pv::getPVSize(const QString& deviceNode)
{
    QString val = getpvField(QStringLiteral("pv_size"), deviceNode);
    return val.isEmpty() ? -1 : val.toLongLong();
}

qint64 lvm2_pv::getPVSize(const QStringList& deviceNodeList)
{
    qint64 sum = 0;
    for (QString deviceNode : deviceNodeList) {
        qint64 pvsize = getPVSize(deviceNode);
        if (pvsize < 0) {
            sum = -1;
            break;
        }
        sum += pvsize;
    }
    return sum;
}

qint64 lvm2_pv::getPESize(const QString& deviceNode)
{
    QString val = getpvField(QStringLiteral("vg_extent_size"), deviceNode);
    return val.isEmpty() ? -1 : val.toLongLong();
}

QString  lvm2_pv::getpvField(const QString& fieldname, const QString& deviceNode)
{
    QStringList args = { QStringLiteral("pvs"),
                    QStringLiteral("--foreign"),
                    QStringLiteral("--readonly"),
                    QStringLiteral("--noheadings"),
                    QStringLiteral("--units"),
                    QStringLiteral("B"),
                    QStringLiteral("--nosuffix"),
                    QStringLiteral("--options"),
                    fieldname };
    if (!deviceNode.isEmpty()) {
        args << deviceNode;
    }
    ExternalCommand cmd(QStringLiteral("lvm"), args);
    if (cmd.run(-1) && cmd.exitCode() == 0) {
        return cmd.output().trimmed();
    }
    return QString();
}

QString lvm2_pv::getVGName(const QString& deviceNode)
{
    return getpvField(QStringLiteral("vg_name"), deviceNode);
}

const QStringList lvm2_pv::getFreePV()
{
    QStringList rlist;

    QString output = getpvField(QStringLiteral("pv_name"));
    const QStringList pvList = output.split(QStringLiteral("\n"), QString::SkipEmptyParts);
    for (QString pvnode : pvList) {
        if (!isUsed(pvnode.trimmed())) {
            rlist.append(pvnode.trimmed());
        }
    }

    return rlist;
}

bool lvm2_pv::isUsed(const QString& deviceNode)
{
    QString output = getpvField(QStringLiteral("pv_in_use"), deviceNode.trimmed());
    if (output.trimmed() == QStringLiteral("used")) {
        return true;
    }
    return false;
}

}
